#!/usr/bin/ruby

require 'bfs_brute_force'

# Puzzle:
#
# Swap black and white bishops, following standard chess movement
# rules, except that bishops may not move to a square that would allow
# them to be captured by an enemy bishop (they may not put themselves
# in "check").
#
# This is the "four bishops" puzzle from an old video game, The 7th Guest.
#
# Initial Board layout:
#
#       +----+----+----+----+----+
#     4 | BB |    |    |    | WB |
#       +----+----+----+----+----+
#     3 | BB |    |    |    | WB |
#       +----+----+----+----+----+
#     2 | BB |    |    |    | WB |
#       +----+----+----+----+----+
#     1 | BB |    |    |    | WB |
#       +----+----+----+----+----+
#         a    b    c    d    e
#
#     BB = Black Bishop
#     WB = White Bishop

class FourBishopsState < BfsBruteForce::State
  # Legal moves: from_position => [to_position, ...]
  @@moves = {
    :a1 => [:b2, :c3, :d4],
    :a2 => [:b1, :b3, :c4],
    :a3 => [:b2, :b4, :c1],
    :a4 => [:b3, :c2, :d1],
    :b1 => [:a2, :c2, :d3, :e4],
    :b2 => [:a1, :a3, :c1, :c3, :d4],
    :b3 => [:a2, :a4, :c2, :c4, :d1],
    :b4 => [:a3, :c3, :d2, :e1],
    :c1 => [:a3, :b2, :d2, :e3],
    :c2 => [:a4, :b1, :b3, :d1, :d3, :e4],
    :c3 => [:a1, :b2, :b4, :d2, :d4, :e1],
    :c4 => [:a2, :b3, :d3, :e2],
    :d1 => [:a4, :b3, :c2, :e2],
    :d2 => [:b4, :c1, :c3, :e1, :e3],
    :d3 => [:b1, :c2, :c4, :c4, :e2],
    :d4 => [:a1, :b2, :c3, :e3],
    :e1 => [:b4, :c3, :d2],
    :e2 => [:c4, :d1, :d3],
    :e3 => [:c1, :d2, :d4],
    :e4 => [:b1, :c2, :d3]
  }

  def initialize(bishops = nil)
    # State of the board: position => bishop
    @bishops = bishops || {
      :a1 => :BB, :a2 => :BB, :a3 => :BB, :a4 => :BB,
      :e1 => :WB, :e2 => :WB, :e3 => :WB, :e4 => :WB
    }
  end

  # (see BfsBruteForce::State#solved)
  def solved?
    @bishops == {
      :a1 => :WB, :a2 => :WB, :a3 => :WB, :a4 => :WB,
      :e1 => :BB, :e2 => :BB, :e3 => :BB, :e4 => :BB
    }
  end

  # Yield all not previously seen states from the current state
  # (see BfsBruteForce::State#already_seen)
  def next_states(already_seen)
    @bishops.each do |from, bishop|
      enemy   = bishop == :WB ? :BB : :WB
      illegal = @bishops.select {|_, b| b == enemy}.flat_map {|p, _| @@moves[p]}

      @@moves[from].reject do |to|
        # Skip illegal positions
        @bishops[to] or illegal.include?(to)
      end.each do |to|
        new_bishops = @bishops.merge(to => bishop)
        new_bishops.delete from

        if already_seen.add?(new_bishops)
          state = FourBishopsState.new new_bishops
          move  = "Move #{bishop} from #{from} to #{to}\n#{state}"
          yield move, state
        end
      end
    end
  end

  def to_s
    fmt = %Q{
        +----+----+----+----+----+
      4 | %s | %s | %s | %s | %s |
        +----+----+----+----+----+
      3 | %s | %s | %s | %s | %s |
        +----+----+----+----+----+
      2 | %s | %s | %s | %s | %s |
        +----+----+----+----+----+
      1 | %s | %s | %s | %s | %s |
        +----+----+----+----+----+
          a    b    c    d    e
    }
    fmt % [
      :a4, :b4, :c4, :d4, :e4,
      :a3, :b3, :c3, :d3, :e3,
      :a2, :b2, :c2, :d2, :e2,
      :a1, :b1, :c1, :d1, :e1
    ].map {|index| @bishops[index] || '  '}
  end
end

solver = BfsBruteForce::Solver.new
moves  = solver.solve(FourBishopsState.new).moves

puts moves

# Running this code will produce the following output:
#
# Move BB from a2 to b3
#
#         +----+----+----+----+----+
#       4 | BB |    |    |    | WB |
#         +----+----+----+----+----+
#       3 | BB | BB |    |    | WB |
#         +----+----+----+----+----+
#       2 |    |    |    |    | WB |
#         +----+----+----+----+----+
#       1 | BB |    |    |    | WB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from a3 to b2
#
#         +----+----+----+----+----+
#       4 | BB |    |    |    | WB |
#         +----+----+----+----+----+
#       3 |    | BB |    |    | WB |
#         +----+----+----+----+----+
#       2 |    | BB |    |    | WB |
#         +----+----+----+----+----+
#       1 | BB |    |    |    | WB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from e1 to b4
#
#         +----+----+----+----+----+
#       4 | BB | WB |    |    | WB |
#         +----+----+----+----+----+
#       3 |    | BB |    |    | WB |
#         +----+----+----+----+----+
#       2 |    | BB |    |    | WB |
#         +----+----+----+----+----+
#       1 | BB |    |    |    |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from e2 to d3
#
#         +----+----+----+----+----+
#       4 | BB | WB |    |    | WB |
#         +----+----+----+----+----+
#       3 |    | BB |    | WB | WB |
#         +----+----+----+----+----+
#       2 |    | BB |    |    |    |
#         +----+----+----+----+----+
#       1 | BB |    |    |    |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from a4 to d1
#
#         +----+----+----+----+----+
#       4 |    | WB |    |    | WB |
#         +----+----+----+----+----+
#       3 |    | BB |    | WB | WB |
#         +----+----+----+----+----+
#       2 |    | BB |    |    |    |
#         +----+----+----+----+----+
#       1 | BB |    |    | BB |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from e3 to d2
#
#         +----+----+----+----+----+
#       4 |    | WB |    |    | WB |
#         +----+----+----+----+----+
#       3 |    | BB |    | WB |    |
#         +----+----+----+----+----+
#       2 |    | BB |    | WB |    |
#         +----+----+----+----+----+
#       1 | BB |    |    | BB |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from b2 to d4
#
#         +----+----+----+----+----+
#       4 |    | WB |    | BB | WB |
#         +----+----+----+----+----+
#       3 |    | BB |    | WB |    |
#         +----+----+----+----+----+
#       2 |    |    |    | WB |    |
#         +----+----+----+----+----+
#       1 | BB |    |    | BB |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from b4 to a3
#
#         +----+----+----+----+----+
#       4 |    |    |    | BB | WB |
#         +----+----+----+----+----+
#       3 | WB | BB |    | WB |    |
#         +----+----+----+----+----+
#       2 |    |    |    | WB |    |
#         +----+----+----+----+----+
#       1 | BB |    |    | BB |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from d3 to b1
#
#         +----+----+----+----+----+
#       4 |    |    |    | BB | WB |
#         +----+----+----+----+----+
#       3 | WB | BB |    |    |    |
#         +----+----+----+----+----+
#       2 |    |    |    | WB |    |
#         +----+----+----+----+----+
#       1 | BB | WB |    | BB |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from b3 to c4
#
#         +----+----+----+----+----+
#       4 |    |    | BB | BB | WB |
#         +----+----+----+----+----+
#       3 | WB |    |    |    |    |
#         +----+----+----+----+----+
#       2 |    |    |    | WB |    |
#         +----+----+----+----+----+
#       1 | BB | WB |    | BB |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from d1 to e2
#
#         +----+----+----+----+----+
#       4 |    |    | BB | BB | WB |
#         +----+----+----+----+----+
#       3 | WB |    |    |    |    |
#         +----+----+----+----+----+
#       2 |    |    |    | WB | BB |
#         +----+----+----+----+----+
#       1 | BB | WB |    |    |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from e4 to c2
#
#         +----+----+----+----+----+
#       4 |    |    | BB | BB |    |
#         +----+----+----+----+----+
#       3 | WB |    |    |    |    |
#         +----+----+----+----+----+
#       2 |    |    | WB | WB | BB |
#         +----+----+----+----+----+
#       1 | BB | WB |    |    |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from d2 to c1
#
#         +----+----+----+----+----+
#       4 |    |    | BB | BB |    |
#         +----+----+----+----+----+
#       3 | WB |    |    |    |    |
#         +----+----+----+----+----+
#       2 |    |    | WB |    | BB |
#         +----+----+----+----+----+
#       1 | BB | WB | WB |    |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from a1 to c3
#
#         +----+----+----+----+----+
#       4 |    |    | BB | BB |    |
#         +----+----+----+----+----+
#       3 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       2 |    |    | WB |    | BB |
#         +----+----+----+----+----+
#       1 |    | WB | WB |    |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from c2 to a4
#
#         +----+----+----+----+----+
#       4 | WB |    | BB | BB |    |
#         +----+----+----+----+----+
#       3 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       2 |    |    |    |    | BB |
#         +----+----+----+----+----+
#       1 |    | WB | WB |    |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from b1 to c2
#
#         +----+----+----+----+----+
#       4 | WB |    | BB | BB |    |
#         +----+----+----+----+----+
#       3 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       2 |    |    | WB |    | BB |
#         +----+----+----+----+----+
#       1 |    |    | WB |    |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from c4 to a2
#
#         +----+----+----+----+----+
#       4 | WB |    |    | BB |    |
#         +----+----+----+----+----+
#       3 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       2 | BB |    | WB |    | BB |
#         +----+----+----+----+----+
#       1 |    |    | WB |    |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from e2 to c4
#
#         +----+----+----+----+----+
#       4 | WB |    | BB | BB |    |
#         +----+----+----+----+----+
#       3 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       2 | BB |    | WB |    |    |
#         +----+----+----+----+----+
#       1 |    |    | WB |    |    |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from c3 to e1
#
#         +----+----+----+----+----+
#       4 | WB |    | BB | BB |    |
#         +----+----+----+----+----+
#       3 | WB |    |    |    |    |
#         +----+----+----+----+----+
#       2 | BB |    | WB |    |    |
#         +----+----+----+----+----+
#       1 |    |    | WB |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from d4 to c3
#
#         +----+----+----+----+----+
#       4 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       3 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       2 | BB |    | WB |    |    |
#         +----+----+----+----+----+
#       1 |    |    | WB |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from c1 to e3
#
#         +----+----+----+----+----+
#       4 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       3 | WB |    | BB |    | WB |
#         +----+----+----+----+----+
#       2 | BB |    | WB |    |    |
#         +----+----+----+----+----+
#       1 |    |    |    |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from a3 to c1
#
#         +----+----+----+----+----+
#       4 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       3 |    |    | BB |    | WB |
#         +----+----+----+----+----+
#       2 | BB |    | WB |    |    |
#         +----+----+----+----+----+
#       1 |    |    | WB |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from c2 to d1
#
#         +----+----+----+----+----+
#       4 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       3 |    |    | BB |    | WB |
#         +----+----+----+----+----+
#       2 | BB |    |    |    |    |
#         +----+----+----+----+----+
#       1 |    |    | WB | WB | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from a2 to b1
#
#         +----+----+----+----+----+
#       4 | WB |    | BB |    |    |
#         +----+----+----+----+----+
#       3 |    |    | BB |    | WB |
#         +----+----+----+----+----+
#       2 |    |    |    |    |    |
#         +----+----+----+----+----+
#       1 |    | BB | WB | WB | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from c4 to d3
#
#         +----+----+----+----+----+
#       4 | WB |    |    |    |    |
#         +----+----+----+----+----+
#       3 |    |    | BB | BB | WB |
#         +----+----+----+----+----+
#       2 |    |    |    |    |    |
#         +----+----+----+----+----+
#       1 |    | BB | WB | WB | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from c3 to b4
#
#         +----+----+----+----+----+
#       4 | WB | BB |    |    |    |
#         +----+----+----+----+----+
#       3 |    |    |    | BB | WB |
#         +----+----+----+----+----+
#       2 |    |    |    |    |    |
#         +----+----+----+----+----+
#       1 |    | BB | WB | WB | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from e3 to d4
#
#         +----+----+----+----+----+
#       4 | WB | BB |    | WB |    |
#         +----+----+----+----+----+
#       3 |    |    |    | BB |    |
#         +----+----+----+----+----+
#       2 |    |    |    |    |    |
#         +----+----+----+----+----+
#       1 |    | BB | WB | WB | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from c1 to b2
#
#         +----+----+----+----+----+
#       4 | WB | BB |    | WB |    |
#         +----+----+----+----+----+
#       3 |    |    |    | BB |    |
#         +----+----+----+----+----+
#       2 |    | WB |    |    |    |
#         +----+----+----+----+----+
#       1 |    | BB |    | WB | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from d1 to b3
#
#         +----+----+----+----+----+
#       4 | WB | BB |    | WB |    |
#         +----+----+----+----+----+
#       3 |    | WB |    | BB |    |
#         +----+----+----+----+----+
#       2 |    | WB |    |    |    |
#         +----+----+----+----+----+
#       1 |    | BB |    |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from b1 to e4
#
#         +----+----+----+----+----+
#       4 | WB | BB |    | WB | BB |
#         +----+----+----+----+----+
#       3 |    | WB |    | BB |    |
#         +----+----+----+----+----+
#       2 |    | WB |    |    |    |
#         +----+----+----+----+----+
#       1 |    |    |    |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from d3 to e2
#
#         +----+----+----+----+----+
#       4 | WB | BB |    | WB | BB |
#         +----+----+----+----+----+
#       3 |    | WB |    |    |    |
#         +----+----+----+----+----+
#       2 |    | WB |    |    | BB |
#         +----+----+----+----+----+
#       1 |    |    |    |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from b4 to d2
#
#         +----+----+----+----+----+
#       4 | WB |    |    | WB | BB |
#         +----+----+----+----+----+
#       3 |    | WB |    |    |    |
#         +----+----+----+----+----+
#       2 |    | WB |    | BB | BB |
#         +----+----+----+----+----+
#       1 |    |    |    |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from d4 to a1
#
#         +----+----+----+----+----+
#       4 | WB |    |    |    | BB |
#         +----+----+----+----+----+
#       3 |    | WB |    |    |    |
#         +----+----+----+----+----+
#       2 |    | WB |    | BB | BB |
#         +----+----+----+----+----+
#       1 | WB |    |    |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from b2 to a3
#
#         +----+----+----+----+----+
#       4 | WB |    |    |    | BB |
#         +----+----+----+----+----+
#       3 | WB | WB |    |    |    |
#         +----+----+----+----+----+
#       2 |    |    |    | BB | BB |
#         +----+----+----+----+----+
#       1 | WB |    |    |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move WB from b3 to a2
#
#         +----+----+----+----+----+
#       4 | WB |    |    |    | BB |
#         +----+----+----+----+----+
#       3 | WB |    |    |    |    |
#         +----+----+----+----+----+
#       2 | WB |    |    | BB | BB |
#         +----+----+----+----+----+
#       1 | WB |    |    |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
# Move BB from d2 to e3
#
#         +----+----+----+----+----+
#       4 | WB |    |    |    | BB |
#         +----+----+----+----+----+
#       3 | WB |    |    |    | BB |
#         +----+----+----+----+----+
#       2 | WB |    |    |    | BB |
#         +----+----+----+----+----+
#       1 | WB |    |    |    | BB |
#         +----+----+----+----+----+
#           a    b    c    d    e
#
