#!/usr/bin/ruby

require 'bfs_brute_force'

# Puzzle:
#
# Swap the white and black knights, using standard chess moves.
# This is the "two knights" puzzle from an old video game, The 11th Hour.
#
# Initial board layout:
#
#       +----+
#     4 | BK |
#       +----+----+----+----+
#     3 |    |    |    | WK |
#       +----+----+----+----+
#     2 | BK | WK |    |
#       +----+----+----+
#     1 |    |    |
#       +----+----+
#         a    b    c    d
#
#     BK = Black Knight
#     WK = White Knight

class KnightsState < BfsBruteForce::State
  # Legal moves: from_position => [to_position, ...]
  @@moves = {
    :a1 => [:b3, :c2],
    :a2 => [:c3],
    :a3 => [:b1, :c2],
    :a4 => [:b2, :c3],
    :b1 => [:a3, :c3],
    :b2 => [:a4, :d3],
    :b3 => [:a1],
    :c2 => [:a1, :a3],
    :c3 => [:a2, :a4, :b1],
    :d3 => [:b2]
  }

  def initialize(knights = nil)
    # State of the board: position => knight
    @knights = knights || {
      :a2 => :BK,
      :a4 => :BK,
      :b2 => :WK,
      :d3 => :WK
    }
  end

  # (see BfsBruteForce::State#solved)
  def solved?
    @knights == {
      :a2 => :WK,
      :a4 => :WK,
      :b2 => :BK,
      :d3 => :BK
    }
  end

  # Yield all not previously seen states from the current state
  # (see BfsBruteForce::State#next_states)
  def next_states(already_seen)
    @knights.each do |from, knight|
      @@moves[from].reject do |to|
        # Skip occupied positions
        @knights[to]
      end.each do |to|
        new_knights = @knights.merge(to => knight)
        new_knights.delete from

        if already_seen.add?(new_knights)
          state = KnightsState.new new_knights
          move  = "Move #{knight} from #{from} to #{to}\n#{state}"
          yield move, state
        end
      end
    end
  end

  def to_s
    fmt = %q{
        +----+
      4 | %s |
        +----+----+----+----+
      3 | %s | %s | %s | %s |
        +----+----+----+----+
      2 | %s | %s | %s |
        +----+----+----+
      1 | %s | %s |
        +----+----+
          a    b    c    d
    }

    fmt % [:a4, :a3, :b3, :c3, :d3, :a2, :b2, :c2, :a1, :b1].map do |index|
      @knights[index] || '  '
    end
  end
end

solver = BfsBruteForce::Solver.new
moves  = solver.solve(KnightsState.new).moves

puts moves

# Running this code will produce the following output:
#
# Move BK from a2 to c3
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    | BK | WK |
#         +----+----+----+----+
#       2 |    | WK |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move BK from c3 to b1
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    |    | WK |
#         +----+----+----+----+
#       2 |    | WK |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move BK from a4 to c3
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 |    |    | BK | WK |
#         +----+----+----+----+
#       2 |    | WK |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move WK from b2 to a4
#
#         +----+
#       4 | WK |
#         +----+----+----+----+
#       3 |    |    | BK | WK |
#         +----+----+----+----+
#       2 |    |    |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move BK from b1 to a3
#
#         +----+
#       4 | WK |
#         +----+----+----+----+
#       3 | BK |    | BK | WK |
#         +----+----+----+----+
#       2 |    |    |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move BK from c3 to b1
#
#         +----+
#       4 | WK |
#         +----+----+----+----+
#       3 | BK |    |    | WK |
#         +----+----+----+----+
#       2 |    |    |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move WK from a4 to c3
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 | BK |    | WK | WK |
#         +----+----+----+----+
#       2 |    |    |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move WK from c3 to a2
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 | BK |    |    | WK |
#         +----+----+----+----+
#       2 | WK |    |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move BK from b1 to c3
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 | BK |    | BK | WK |
#         +----+----+----+----+
#       2 | WK |    |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move BK from a3 to b1
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 |    |    | BK | WK |
#         +----+----+----+----+
#       2 | WK |    |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move BK from c3 to a4
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    |    | WK |
#         +----+----+----+----+
#       2 | WK |    |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move BK from b1 to c3
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    | BK | WK |
#         +----+----+----+----+
#       2 | WK |    |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move BK from a4 to b2
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 |    |    | BK | WK |
#         +----+----+----+----+
#       2 | WK | BK |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move BK from c3 to a4
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    |    | WK |
#         +----+----+----+----+
#       2 | WK | BK |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move WK from a2 to c3
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    | WK | WK |
#         +----+----+----+----+
#       2 |    | BK |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move WK from c3 to b1
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    |    | WK |
#         +----+----+----+----+
#       2 |    | BK |    |
#         +----+----+----+
#       1 |    | WK |
#         +----+----+
#           a    b    c    d
#
# Move BK from a4 to c3
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 |    |    | BK | WK |
#         +----+----+----+----+
#       2 |    | BK |    |
#         +----+----+----+
#       1 |    | WK |
#         +----+----+
#           a    b    c    d
#
# Move BK from b2 to a4
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    | BK | WK |
#         +----+----+----+----+
#       2 |    |    |    |
#         +----+----+----+
#       1 |    | WK |
#         +----+----+
#           a    b    c    d
#
# Move WK from d3 to b2
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    | BK |    |
#         +----+----+----+----+
#       2 |    | WK |    |
#         +----+----+----+
#       1 |    | WK |
#         +----+----+
#           a    b    c    d
#
# Move WK from b1 to a3
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 | WK |    | BK |    |
#         +----+----+----+----+
#       2 |    | WK |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move BK from c3 to b1
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 | WK |    |    |    |
#         +----+----+----+----+
#       2 |    | WK |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move BK from a4 to c3
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 | WK |    | BK |    |
#         +----+----+----+----+
#       2 |    | WK |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move WK from b2 to a4
#
#         +----+
#       4 | WK |
#         +----+----+----+----+
#       3 | WK |    | BK |    |
#         +----+----+----+----+
#       2 |    |    |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move WK from a3 to c2
#
#         +----+
#       4 | WK |
#         +----+----+----+----+
#       3 |    |    | BK |    |
#         +----+----+----+----+
#       2 |    |    | WK |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move BK from b1 to a3
#
#         +----+
#       4 | WK |
#         +----+----+----+----+
#       3 | BK |    | BK |    |
#         +----+----+----+----+
#       2 |    |    | WK |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move BK from c3 to b1
#
#         +----+
#       4 | WK |
#         +----+----+----+----+
#       3 | BK |    |    |    |
#         +----+----+----+----+
#       2 |    |    | WK |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move WK from a4 to c3
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 | BK |    | WK |    |
#         +----+----+----+----+
#       2 |    |    | WK |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move WK from c3 to a2
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 | BK |    |    |    |
#         +----+----+----+----+
#       2 | WK |    | WK |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move BK from b1 to c3
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 | BK |    | BK |    |
#         +----+----+----+----+
#       2 | WK |    | WK |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move BK from a3 to b1
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 |    |    | BK |    |
#         +----+----+----+----+
#       2 | WK |    | WK |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move WK from c2 to a3
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 | WK |    | BK |    |
#         +----+----+----+----+
#       2 | WK |    |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move BK from c3 to a4
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 | WK |    |    |    |
#         +----+----+----+----+
#       2 | WK |    |    |
#         +----+----+----+
#       1 |    | BK |
#         +----+----+
#           a    b    c    d
#
# Move BK from b1 to c3
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 | WK |    | BK |    |
#         +----+----+----+----+
#       2 | WK |    |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move WK from a3 to b1
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    | BK |    |
#         +----+----+----+----+
#       2 | WK |    |    |
#         +----+----+----+
#       1 |    | WK |
#         +----+----+
#           a    b    c    d
#
# Move BK from a4 to b2
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 |    |    | BK |    |
#         +----+----+----+----+
#       2 | WK | BK |    |
#         +----+----+----+
#       1 |    | WK |
#         +----+----+
#           a    b    c    d
#
# Move BK from c3 to a4
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    |    |    |
#         +----+----+----+----+
#       2 | WK | BK |    |
#         +----+----+----+
#       1 |    | WK |
#         +----+----+
#           a    b    c    d
#
# Move WK from b1 to c3
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    | WK |    |
#         +----+----+----+----+
#       2 | WK | BK |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move BK from b2 to d3
#
#         +----+
#       4 | BK |
#         +----+----+----+----+
#       3 |    |    | WK | BK |
#         +----+----+----+----+
#       2 | WK |    |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move BK from a4 to b2
#
#         +----+
#       4 |    |
#         +----+----+----+----+
#       3 |    |    | WK | BK |
#         +----+----+----+----+
#       2 | WK | BK |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
# Move WK from c3 to a4
#
#         +----+
#       4 | WK |
#         +----+----+----+----+
#       3 |    |    |    | BK |
#         +----+----+----+----+
#       2 | WK | BK |    |
#         +----+----+----+
#       1 |    |    |
#         +----+----+
#           a    b    c    d
#
